@@ -1,20 +1,22 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import datetime
 from decimal import Decimal
-from trytond.model import ModelView, ModelSQL, fields
-from trytond.wizard import Wizard, StateView, Button, StateAction
+
+from trytond.model import (ModelView, ModelSQL, MatchMixin, ValueMixin, fields,
+    sequence_ordered)
 from trytond import backend
-from trytond.pyson import Not, Bool, Eval, Equal, PYSONEncoder, Date
+from trytond.pyson import Eval, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
+from trytond.tools import grouped_slice
+from trytond.tools.multivalue import migrate_property
 
-__all__ = ['Location', 'Party', 'ProductsByLocationsStart',
-    'ProductsByLocations']
-__metaclass__ = PoolMeta
+__all__ = ['Location', 'Party', 'PartyLocation', 'ProductsByLocationsContext',
+    'LocationLeadTime']
 
 STATES = {
-    'readonly': Not(Bool(Eval('active'))),
+    'readonly': ~Eval('active'),
 }
 DEPENDS = ['active']
 
@@ -29,8 +31,8 @@ class Location(ModelSQL, ModelView):
     active = fields.Boolean('Active', select=True)
     address = fields.Many2One("party.address", "Address",
         states={
-            'invisible': Not(Equal(Eval('type'), 'warehouse')),
-            'readonly': Not(Bool(Eval('active'))),
+            'invisible': Eval('type') != 'warehouse',
+            'readonly': ~Eval('active'),
             },
         depends=['type', 'active'])
     type = fields.Selection([
@@ -40,18 +42,25 @@ class Location(ModelSQL, ModelView):
         ('warehouse', 'Warehouse'),
         ('storage', 'Storage'),
         ('production', 'Production'),
+        ('drop', 'Drop'),
         ('view', 'View'),
         ], 'Location type', states=STATES, depends=DEPENDS)
     parent = fields.Many2One("stock.location", "Parent", select=True,
-            left="left", right="right")
+        left="left", right="right",
+        states={
+            'invisible': Eval('type') == 'warehouse',
+            },
+        depends=['type'])
     left = fields.Integer('Left', required=True, select=True)
     right = fields.Integer('Right', required=True, select=True)
     childs = fields.One2Many("stock.location", "parent", "Children")
+    warehouse = fields.Function(fields.Many2One('stock.location', 'Warehouse'),
+        'get_warehouse')
     input_location = fields.Many2One(
         "stock.location", "Input", states={
-            'invisible': Not(Equal(Eval('type'), 'warehouse')),
-            'readonly': Not(Bool(Eval('active'))),
-            'required': Equal(Eval('type'), 'warehouse'),
+            'invisible': Eval('type') != 'warehouse',
+            'readonly': ~Eval('active'),
+            'required': Eval('type') == 'warehouse',
             },
         domain=[
             ('type', '=', 'storage'),
@@ -63,9 +72,9 @@ class Location(ModelSQL, ModelView):
         depends=['type', 'active', 'id'])
     output_location = fields.Many2One(
         "stock.location", "Output", states={
-            'invisible': Not(Equal(Eval('type'), 'warehouse')),
-            'readonly': Not(Bool(Eval('active'))),
-            'required': Equal(Eval('type'), 'warehouse'),
+            'invisible': Eval('type') != 'warehouse',
+            'readonly': ~Eval('active'),
+            'required': Eval('type') == 'warehouse',
         },
         domain=[
             ('type', '=', 'storage'),
@@ -75,16 +84,27 @@ class Location(ModelSQL, ModelView):
         depends=['type', 'active', 'id'])
     storage_location = fields.Many2One(
         "stock.location", "Storage", states={
-            'invisible': Not(Equal(Eval('type'), 'warehouse')),
-            'readonly': Not(Bool(Eval('active'))),
-            'required': Equal(Eval('type'), 'warehouse'),
+            'invisible': Eval('type') != 'warehouse',
+            'readonly': ~Eval('active'),
+            'required': Eval('type') == 'warehouse',
         },
         domain=[
-            ('type', '=', 'storage'),
+            ('type', 'in', ['storage', 'view']),
             ['OR',
                 ('parent', 'child_of', [Eval('id')]),
                 ('parent', '=', None)]],
         depends=['type', 'active', 'id'])
+    picking_location = fields.Many2One(
+        'stock.location', 'Picking', states={
+            'invisible': Eval('type') != 'warehouse',
+            'readonly': ~Eval('active'),
+            },
+        domain=[
+            ('type', '=', 'storage'),
+            ('parent', 'child_of', [Eval('storage_location', -1)]),
+            ],
+        depends=['type', 'active', 'storage_location'],
+        help='If empty the Storage is used')
     quantity = fields.Function(fields.Float('Quantity'), 'get_quantity')
     forecast_quantity = fields.Function(fields.Float('Forecast Quantity'),
             'get_quantity')
@@ -103,13 +123,47 @@ class Location(ModelSQL, ModelView):
                     'child of warehouse "%(warehouse)s".'),
                 })
 
+        parent_domain = []
+        childs_domain = []
+        childs_mapping = cls._childs_domain()
+        for type_, allowed_parents in cls._parent_domain().iteritems():
+            parent_domain.append(If(Eval('type') == type_,
+                    ('type', 'in', allowed_parents), ()))
+            childs_domain.append(If(Eval('type') == type_,
+                    ('type', 'in', childs_mapping[type_]), ()))
+        cls.parent.domain = parent_domain
+        cls.childs.domain = childs_domain
+        cls.childs.depends.append('type')
+
+    @classmethod
+    def _parent_domain(cls):
+        '''Returns a dict with location types as keys and a list of allowed
+        parent location types as values'''
+        return {
+            'customer': ['customer'],
+            'supplier': ['supplier'],
+            'production': ['production'],
+            'lost_found': ['lost_found'],
+            'view': ['warehouse', 'view', 'storage'],
+            'storage': ['warehouse', 'view', 'storage'],
+            'warehouse': [''],
+            }
+
+    @classmethod
+    def _childs_domain(cls):
+        childs_domain = {}
+        for type_, allowed_parents in cls._parent_domain().iteritems():
+            for parent in allowed_parents:
+                childs_domain.setdefault(parent, [])
+                childs_domain[parent].append(type_)
+        return childs_domain
+
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
         super(Location, cls).__register__(module_name)
-        cursor = Transaction().cursor
 
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action(['left', 'right'], 'add')
 
     @classmethod
@@ -123,12 +177,19 @@ class Location(ModelSQL, ModelView):
         """ Check locations with moves have types compatible with moves. """
         invalid_move_types = ['warehouse', 'view']
         Move = Pool().get('stock.move')
-        if (self.type in invalid_move_types
-                and Move.search(['OR',
-                        ('to_location', '=', self.id),
-                        ('from_location', '=', self.id),
-                        ])):
-            self.raise_user_error('invalid_type_for_moves', (self.rec_name,))
+        if self.type in invalid_move_types:
+            # Use root to compute for all companies
+            with Transaction().set_user(0):
+                moves = Move.search([
+                        ['OR',
+                            ('to_location', '=', self.id),
+                            ('from_location', '=', self.id),
+                            ],
+                        ('state', 'not in', ['staging', 'draft']),
+                        ])
+            if moves:
+                self.raise_user_error(
+                    'invalid_type_for_moves', (self.rec_name,))
 
     @staticmethod
     def default_active():
@@ -150,14 +211,26 @@ class Location(ModelSQL, ModelView):
     def check_xml_record(self, records, values):
         return True
 
+    def get_warehouse(self, name):
+        # Order by descending left to get the first one in the tree
+        with Transaction().set_context(active_test=False):
+            locations = self.search([
+                    ('parent', 'parent_of', [self.id]),
+                    ('type', '=', 'warehouse'),
+                    ], order=[('left', 'DESC')])
+        if locations:
+            return locations[0].id
+
     @classmethod
     def search_rec_name(cls, name, clause):
-        locations = cls.search([
-                ('code', '=', clause[2]),
-                ], order=[])
-        if locations:
-            return [('id', 'in', [l.id for l in locations])]
-        return [(cls._rec_name,) + tuple(clause[1:])]
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        return [bool_op,
+            (cls._rec_name,) + tuple(clause[1:]),
+            ('code',) + tuple(clause[1:]),
+            ]
 
     @classmethod
     def get_quantity(cls, locations, name):
@@ -178,7 +251,7 @@ class Location(ModelSQL, ModelView):
 
         context = {}
         if (name == 'quantity'
-                and Transaction().context.get('stock_date_end') >
+                and Transaction().context.get('stock_date_end', datetime.date.max) >
                 Date_.today()):
             context['stock_date_end'] = Date_.today()
 
@@ -187,14 +260,17 @@ class Location(ModelSQL, ModelView):
             if not Transaction().context.get('stock_date_end'):
                 context['stock_date_end'] = datetime.date.max
 
-        location_ids = [l.id for l in locations]
         product_id = Transaction().context['product']
-        with Transaction().set_context(context):
-            pbl = Product.products_by_location(location_ids=location_ids,
-                product_ids=[product_id], with_childs=True)
+        pbl = {}
+        for sub_locations in grouped_slice(locations):
+            location_ids = [l.id for l in sub_locations]
+            with Transaction().set_context(context):
+                pbl.update(Product.products_by_location(
+                        location_ids=location_ids, product_ids=[product_id],
+                        with_childs=True))
 
-        return dict((loc, pbl.get((loc, product_id), 0))
-            for loc in location_ids)
+        return dict((loc.id, pbl.get((loc.id, product_id), 0))
+            for loc in locations)
 
     @classmethod
     def get_cost_value(cls, locations, name):
@@ -205,7 +281,9 @@ class Location(ModelSQL, ModelView):
             return dict((l.id, None) for l in locations)
         cost_values, context = {}, {}
         if 'stock_date_end' in trans_context:
-            context['_datetime'] = trans_context['stock_date_end']
+            # Use the last cost_price of the day
+            context['_datetime'] = datetime.datetime.combine(
+                trans_context['stock_date_end'], datetime.time.max)
         with Transaction().set_context(context):
             product = Product(product_id)
             for location in locations:
@@ -223,6 +301,7 @@ class Location(ModelSQL, ModelView):
         Set the parent of child location of warehouse if not set
         '''
         to_update = set()
+        to_save = []
         for location in locations:
             if location.type == 'warehouse':
                 if not location.input_location.parent:
@@ -232,10 +311,11 @@ class Location(ModelSQL, ModelView):
                 if not location.storage_location.parent:
                     to_update.add(location.storage_location)
                 if to_update:
-                    cls.write(list(to_update), {
-                        'parent': location.id,
-                        })
+                    for child_location in to_update:
+                        child_location.parent = location
+                        to_save.append(child_location)
                     to_update.clear()
+        cls.save(to_save)
 
     @classmethod
     def create(cls, vlist):
@@ -244,8 +324,9 @@ class Location(ModelSQL, ModelView):
         return locations
 
     @classmethod
-    def write(cls, locations, vals):
-        super(Location, cls).write(locations, vals)
+    def write(cls, *args):
+        super(Location, cls).write(*args)
+        locations = sum(args[::2], [])
         cls._set_warehouse_parent(locations)
 
         ids = [l.id for l in locations]
@@ -278,9 +359,6 @@ class Location(ModelSQL, ModelView):
         if default is None:
             default = {}
 
-        default['left'] = 0
-        default['right'] = 0
-
         res = []
         for location in locations:
             if location.type == 'warehouse':
@@ -325,65 +403,127 @@ class Location(ModelSQL, ModelView):
         return res
 
 
+supplier_location = fields.Many2One(
+    'stock.location', "Supplier Location", domain=[('type', '=', 'supplier')],
+    help='The default source location when receiving products from the party.')
+customer_location = fields.Many2One(
+    'stock.location', "Customer Location", domain=[('type', '=', 'customer')],
+    help='The default destination location when sending products to the party.'
+    )
+
+
 class Party:
+    __metaclass__ = PoolMeta
     __name__ = 'party.party'
-    supplier_location = fields.Property(fields.Many2One('stock.location',
-        'Supplier Location', domain=[('type', '=', 'supplier')],
-        help='The default source location when receiving products from the '
-        'party.'))
-    customer_location = fields.Property(fields.Many2One('stock.location',
-        'Customer Location', domain=[('type', '=', 'customer')],
-        help='The default destination location when sending products to the '
-        'party.'))
+    supplier_location = fields.MultiValue(supplier_location)
+    customer_location = fields.MultiValue(customer_location)
+    locations = fields.One2Many(
+        'party.party.location', 'party', "Locations")
+
+    @classmethod
+    def multivalue_model(cls, field):
+        pool = Pool()
+        if field in {'supplier_location', 'customer_location'}:
+            return pool.get('party.party.location')
+        return super(Party, cls).multivalue_model(field)
+
+    @classmethod
+    def default_supplier_location(cls, **pattern):
+        return cls.multivalue_model(
+            'supplier_location').default_supplier_location()
+
+    @classmethod
+    def default_customer_location(cls, **pattern):
+        return cls.multivalue_model(
+            'customer_location').default_customer_location()
 
 
-class ProductsByLocationsStart(ModelView):
+class PartyLocation(ModelSQL, ValueMixin):
+    "Party Location"
+    __name__ = 'party.party.location'
+    party = fields.Many2One(
+        'party.party', "Party", ondelete='CASCADE', select=True)
+    supplier_location = supplier_location
+    customer_location = customer_location
+
+    @classmethod
+    def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
+        exist = TableHandler.table_exist(cls._table)
+
+        super(PartyLocation, cls).__register__(module_name)
+
+        if not exist:
+            cls._migrate_property([], [], [])
+
+    @classmethod
+    def _migrate_property(cls, field_names, value_names, fields):
+        field_names.extend(['supplier_location', 'customer_location'])
+        value_names.extend(['supplier_location', 'customer_location'])
+        migrate_property(
+            'party.party', field_names, cls, value_names,
+            parent='party', fields=fields)
+
+    @classmethod
+    def default_supplier_location(cls):
+        pool = Pool()
+        ModelData = pool.get('ir.model.data')
+        try:
+            return ModelData.get_id('stock', 'location_supplier')
+        except KeyError:
+            return None
+
+    @classmethod
+    def default_customer_location(cls):
+        pool = Pool()
+        ModelData = pool.get('ir.model.data')
+        try:
+            return ModelData.get_id('stock', 'location_customer')
+        except KeyError:
+            return None
+
+
+class ProductsByLocationsContext(ModelView):
     'Products by Locations'
-    __name__ = 'stock.products_by_locations.start'
+    __name__ = 'stock.products_by_locations.context'
     forecast_date = fields.Date(
         'At Date', help=('Allow to compute expected '
             'stock quantities for this date.\n'
             '* An empty value is an infinite date in the future.\n'
             '* A date in the past will provide historical values.'))
+    stock_date_end = fields.Function(fields.Date('At Date'),
+        'on_change_with_stock_date_end')
 
     @staticmethod
     def default_forecast_date():
         Date_ = Pool().get('ir.date')
         return Date_.today()
 
+    @fields.depends('forecast_date')
+    def on_change_with_stock_date_end(self, name=None):
+        if self.forecast_date is None:
+            return datetime.date.max
+        return self.forecast_date
 
-class ProductsByLocations(Wizard):
-    'Products by Locations'
-    __name__ = 'stock.products_by_locations'
-    start = StateView('stock.products_by_locations.start',
-        'stock.products_by_locations_start_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Open', 'open', 'tryton-ok', True),
+
+class LocationLeadTime(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
+    'Location Lead Time'
+    __name__ = 'stock.location.lead_time'
+
+    warehouse_from = fields.Many2One('stock.location', 'Warehouse From',
+        ondelete='CASCADE',
+        domain=[
+            ('type', '=', 'warehouse'),
             ])
-    open = StateAction('stock.act_products_by_locations')
+    warehouse_to = fields.Many2One('stock.location', 'Warehouse To',
+        ondelete='CASCADE',
+        domain=[
+            ('type', '=', 'warehouse'),
+            ])
+    lead_time = fields.TimeDelta('Lead Time')
 
-    def do_open(self, action):
-        pool = Pool()
-        Location = pool.get('stock.location')
-        Lang = pool.get('ir.lang')
-
-        context = {}
-        context['locations'] = Transaction().context.get('active_ids')
-        date = self.start.forecast_date or datetime.date.max
-        context['stock_date_end'] = Date(date.year, date.month, date.day)
-        action['pyson_context'] = PYSONEncoder().encode(context)
-
-        locations = Location.browse(context['locations'])
-
-        for code in [Transaction().language, 'en_US']:
-            langs = Lang.search([
-                    ('code', '=', code),
-                    ])
-            if langs:
-                break
-        lang = langs[0]
-        date = Lang.strftime(date, lang.code, lang.date)
-
-        action['name'] += ' - (%s) @ %s' % (
-            ','.join(l.name for l in locations), date)
-        return action, {}
+    @classmethod
+    def get_lead_time(cls, pattern):
+        for record in cls.search([]):
+            if record.match(pattern):
+                return record.lead_time
