@@ -1,28 +1,32 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import datetime
 from decimal import Decimal
-from sql import Literal
+from collections import defaultdict
+
+from sql import Literal, Null
 from sql.aggregate import Max
-from sql.functions import Now
+from sql.functions import CurrentTimestamp
 from sql.conditionals import Coalesce
 
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.wizard import Wizard, StateView, StateAction, Button
-from trytond.pyson import PYSONEncoder, Eval, Or
+from trytond.wizard import Wizard, StateTransition
+from trytond.pyson import Eval, Or
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
+from trytond.tools import grouped_slice
 
+from trytond.modules.product import TemplateFunction
 from .move import StockMixin
 
 __all__ = ['Template', 'Product',
-    'ProductByLocationStart', 'ProductByLocation',
-    'ProductQuantitiesByWarehouse', 'ProductQuantitiesByWarehouseStart',
-    'OpenProductQuantitiesByWarehouse']
-__metaclass__ = PoolMeta
+    'ProductByLocationContext',
+    'ProductQuantitiesByWarehouse', 'ProductQuantitiesByWarehouseContext',
+    'RecomputeCostPrice']
 
 
 class Template:
+    __metaclass__ = PoolMeta
     __name__ = "product.template"
     quantity = fields.Function(fields.Float('Quantity'), 'sum_product')
     forecast_quantity = fields.Function(fields.Float('Forecast Quantity'),
@@ -59,11 +63,9 @@ class Template:
     @classmethod
     def check_no_move(cls, templates, error):
         Move = Pool().get('stock.move')
-        cursor = Transaction().cursor
-        for i in range(0, len(templates), cursor.IN_MAX):
-            sub_ids = [t.id for t in templates[i:i + cursor.IN_MAX]]
+        for sub_templates in grouped_slice(templates):
             moves = Move.search([
-                    ('product.template', 'in', sub_ids),
+                    ('product.template', 'in', [t.id for t in sub_templates]),
                     ],
                 limit=1, order=[])
             if moves:
@@ -71,7 +73,8 @@ class Template:
 
     @classmethod
     def write(cls, *args):
-        if Transaction().user != 0:
+        if (Transaction().user != 0
+                and Transaction().context.get('_check_access')):
             actions = iter(args)
             for templates, values in zip(actions, actions):
                 for field, error in cls._modify_no_move:
@@ -80,8 +83,16 @@ class Template:
                         break
         super(Template, cls).write(*args)
 
+    @classmethod
+    def recompute_cost_price(cls, templates):
+        pool = Pool()
+        Product = pool.get('product.product')
 
-class Product(object, StockMixin):
+        products = [p for t in templates for p in t.products]
+        Product.recompute_cost_price(products)
+
+
+class Product(StockMixin, object):
     __metaclass__ = PoolMeta
     __name__ = "product.product"
     quantity = fields.Function(fields.Float('Quantity'), 'get_quantity',
@@ -107,9 +118,11 @@ class Product(object, StockMixin):
         context = {}
         trans_context = Transaction().context
         if trans_context.get('stock_date_end'):
-            context['_datetime'] = trans_context['stock_date_end']
+            # Use the last cost_price of the day
+            context['_datetime'] = datetime.datetime.combine(
+                trans_context['stock_date_end'], datetime.time.max)
         with Transaction().set_context(context):
-            for product in products:
+            for product in cls.browse(products):
                 # The date could be before the product creation
                 if not isinstance(product.cost_price, Decimal):
                     cost_values[product.id] = None
@@ -170,68 +183,141 @@ class Product(object, StockMixin):
                             del quantities[key]
         return quantities
 
+    @classmethod
+    def recompute_cost_price(cls, products):
+        pool = Pool()
+        Template = pool.get('product.template')
 
-class ProductByLocationStart(ModelView):
+        if not isinstance(cls.cost_price, TemplateFunction):
+            digits = cls.cost_price.digits
+            write = cls.write
+            record = lambda p: p
+        else:
+            digits = Template.cost_price.digits
+            write = Template.write
+            record = lambda p: p.template
+
+        costs = defaultdict(list)
+        for product in products:
+            if product.type == 'service':
+                continue
+            cost = getattr(product,
+                'recompute_cost_price_%s' % product.cost_price_method)()
+            cost = cost.quantize(Decimal(str(10.0 ** -digits[1])))
+            costs[cost].append(record(product))
+
+        if not costs:
+            return
+
+        to_write = []
+        for cost, records in costs.iteritems():
+            to_write.append(records)
+            to_write.append({'cost_price': cost})
+
+        # Enforce check access for account_stock*
+        with Transaction().set_context(_check_access=True):
+            write(*to_write)
+
+    def recompute_cost_price_fixed(self):
+        return self.cost_price
+
+    def recompute_cost_price_average(self):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        Currency = pool.get('currency.currency')
+        Uom = pool.get('product.uom')
+
+        context = Transaction().context
+
+        if not isinstance(self.__class__.cost_price, TemplateFunction):
+            product_clause = ('product', '=', self.id)
+        else:
+            product_clause = ('product.template', '=', self.template.id)
+
+        moves = Move.search([
+                product_clause,
+                ('state', '=', 'done'),
+                ('company', '=', context.get('company')),
+                ['OR',
+                    [
+                        ('to_location.type', '=', 'storage'),
+                        ('from_location.type', '!=', 'storage'),
+                        ],
+                    [
+                        ('from_location.type', '=', 'storage'),
+                        ('to_location.type', '!=', 'storage'),
+                        ],
+                    ],
+                ], order=[('effective_date', 'ASC'), ('id', 'ASC')])
+
+        cost_price = Decimal(0)
+        quantity = 0
+        for move in moves:
+            qty = Uom.compute_qty(move.uom, move.quantity, self.default_uom)
+            qty = Decimal(str(qty))
+            if move.from_location.type == 'storage':
+                qty *= -1
+            if (move.from_location.type in ['supplier', 'production']
+                    or move.to_location.type == 'supplier'):
+                with Transaction().set_context(date=move.effective_date):
+                    unit_price = Currency.compute(
+                        move.currency, move.unit_price,
+                        move.company.currency, round=False)
+                unit_price = Uom.compute_price(move.uom, unit_price,
+                    self.default_uom)
+                if quantity + qty != 0:
+                    cost_price = (
+                        (cost_price * quantity) + (unit_price * qty)
+                        ) / (quantity + qty)
+            quantity += qty
+        return cost_price
+
+
+class ProductByLocationContext(ModelView):
     'Product by Location'
-    __name__ = 'product.by_location.start'
+    __name__ = 'product.by_location.context'
     forecast_date = fields.Date(
         'At Date', help=('Allow to compute expected '
             'stock quantities for this date.\n'
             '* An empty value is an infinite date in the future.\n'
             '* A date in the past will provide historical values.'))
+    stock_date_end = fields.Function(fields.Date('At Date'),
+        'on_change_with_stock_date_end')
 
     @staticmethod
     def default_forecast_date():
         Date = Pool().get('ir.date')
         return Date.today()
 
-
-class ProductByLocation(Wizard):
-    'Product by Location'
-    __name__ = 'product.by_location'
-    start = StateView('product.by_location.start',
-        'stock.product_by_location_start_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Open', 'open', 'tryton-ok', default=True),
-            ])
-    open = StateAction('stock.act_location_quantity_tree')
-
-    def do_open(self, action):
-        pool = Pool()
-        Product = pool.get('product.product')
-        Lang = pool.get('ir.lang')
-
-        context = {}
-        product_id = Transaction().context['active_id']
-        context['product'] = product_id
-        if self.start.forecast_date:
-            context['stock_date_end'] = self.start.forecast_date
-        else:
-            context['stock_date_end'] = datetime.date.max
-        action['pyson_context'] = PYSONEncoder().encode(context)
-        product = Product(product_id)
-
-        for code in [Transaction().language, 'en_US']:
-            langs = Lang.search([
-                    ('code', '=', code),
-                    ])
-            if langs:
-                break
-        lang, = langs
-        date = Lang.strftime(context['stock_date_end'],
-            lang.code, lang.date)
-
-        action['name'] += ' - %s (%s) @ %s' % (product.rec_name,
-            product.default_uom.rec_name, date)
-        return action, {}
+    @fields.depends('forecast_date')
+    def on_change_with_stock_date_end(self, name=None):
+        if self.forecast_date is None:
+            return datetime.date.max
+        return self.forecast_date
 
 
 class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
     'Product Quantities By Warehouse'
     __name__ = 'stock.product_quantities_warehouse'
-    date = fields.Date('Date')
+
+    class _Date(fields.Date):
+        def get(self, ids, model, name, values=None):
+            if values is None:
+                values = {}
+            result = {}
+            for v in values:
+                date = v[name]
+                # SQLite does not convert to date
+                if isinstance(date, str):
+                    date = datetime.date(*map(int, date.split('-', 2)))
+                result[v['id']] = date
+            return result
+
+    date = _Date('Date')
     quantity = fields.Function(fields.Float('Quantity'), 'get_quantity')
 
+    del _Date
+
     @classmethod
     def __setup__(cls):
         super(ProductQuantitiesByWarehouse, cls).__setup__()
@@ -254,14 +340,14 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
         return move.select(
             Max(move.id).as_('id'),
             Literal(0).as_('create_uid'),
-            Now().as_('create_date'),
+            CurrentTimestamp().as_('create_date'),
             Literal(None).as_('write_uid'),
             Literal(None).as_('write_date'),
             date_column,
             where=(move.product == product_id)
             & (move.from_location.in_(warehouse_query)
                 | move.to_location.in_(warehouse_query))
-            & (Coalesce(move.effective_date, move.planned_date) != None),
+            & (Coalesce(move.effective_date, move.planned_date) != Null),
             group_by=(date_column, move.product))
 
     @classmethod
@@ -296,10 +382,11 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
         return dict((l.id, quantities[l.date]) for l in lines)
 
 
-class ProductQuantitiesByWarehouseStart(ModelView):
+class ProductQuantitiesByWarehouseContext(ModelView):
     'Product Quantities By Warehouse'
-    __name__ = 'stock.product_quantities_warehouse.start'
-    warehouse = fields.Many2One('stock.location', 'Warehouse', domain=[
+    __name__ = 'stock.product_quantities_warehouse.context'
+    warehouse = fields.Many2One('stock.location', 'Warehouse', required=True,
+        domain=[
             ('type', '=', 'warehouse'),
             ])
 
@@ -313,23 +400,23 @@ class ProductQuantitiesByWarehouseStart(ModelView):
             return warehouses[0].id
 
 
-class OpenProductQuantitiesByWarehouse(Wizard):
-    'Product Quantities By Warehouse'
-    __name__ = 'stock.product_quantities_warehouse'
-    start = StateView('stock.product_quantities_warehouse.start',
-        'stock.product_quantities_warehouse_start_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Open', 'open_', 'tryton-ok', default=True),
-            ])
-    open_ = StateAction('stock.act_product_quantities_warehouse')
+class RecomputeCostPrice(Wizard):
+    'Recompute Cost Price'
+    __name__ = 'product.recompute_cost_price'
+    start_state = 'recompute'
+    recompute = StateTransition()
 
-    def do_open_(self, action):
-        Date = Pool().get('ir.date')
-        action['pyson_context'] = PYSONEncoder().encode({
-                'product': Transaction().context['active_id'],
-                'warehouse': self.start.warehouse.id,
-                })
-        action['pyson_search_value'] = PYSONEncoder().encode([
-                ('date', '>=', Date.today()),
-                ])
-        return action, {}
+    def transition_recompute(self):
+        pool = Pool()
+        Product = pool.get('product.product')
+        Template = pool.get('product.template')
+
+        context = Transaction().context
+
+        if context['active_model'] == 'product.product':
+            products = Product.browse(context['active_ids'])
+            Product.recompute_cost_price(products)
+        elif context['active_model'] == 'product.template':
+            templates = Template.browse(context['active_ids'])
+            Template.recompute_cost_price(templates)
+        return 'end'
